--[[
Name: DogTag-1.0
Revision: $Rev: 33578 $
Author: Cameron Kenneth Knight (ckknight@gmail.com)
Website: http://www.wowace.com/
Documentation: http://www.wowace.com/wiki/DogTag-1.0
SVN: svn://svn.wowace.com/wowace/trunk/DogTag-1.0/DogTag-1.0
Description: A library to provide a markup syntax for unit frame FontStrings
Dependencies: AceLibrary
License: LGPL v2.1
]]

-------------------------------------------------------------------
-- FOR TAG INFORMATION SEE http://www.wowace.com/wiki/DogTag-1.0 --
-------------------------------------------------------------------

local MAJOR_VERSION = "DogTag-1.0"
local MINOR_VERSION = "$Revision: 33578 $"

-- This ensures the code is only executed if the libary doesn't already exist, or is a newer version
if not AceLibrary then error(MAJOR_VERSION .. " requires AceLibrary.") end
if not AceLibrary:IsNewVersion(MAJOR_VERSION, MINOR_VERSION) then return end

local DogTag = {}

local L = {
	-- short races
	["BE"] = "BE",
	["Dr"] = "Dr",
	["Dw"] = "Dw",
	["Gn"] = "Gn",
	["Hu"] = "Hu",
	["NE"] = "NE",
	["Or"] = "Or",
	["Ta"] = "Ta",
	["Tr"] = "Tr",
	["Ud"] = "Ud",
	
	-- short classes
	["Wr"] = "Wr",
	["Pr"] = "Pr",
	["Ma"] = "Ma",
	["Sh"] = "Sh",
	["Pa"] = "Pa",
	["Wl"] = "Wl",
	["Dr"] = "Dr",
	["Ro"] = "Ro",
	["Hu"] = "Hu",
	
	-- classifications
	["Rare"] = ITEM_QUALITY3_DESC,
	["Rare-Elite"] = ITEM_QUALITY3_DESC .. "-" .. ELITE,
	["Elite"] = ELITE,
	["Boss"] = BOSS,
	-- short classifications
	["r"] = "r",
	["r+"] = "r+",
	["+"] = "+",
	["b"] = "b",
	
	["Dead"] = DEAD,
	["Ghost"] = "Ghost",
	["Offline"] = PLAYER_OFFLINE,
	["Combat"] = "Combat",
	["Resting"] = "Resting",
	["Tapped"] = "Tapped",
	
	["Rage"] = RAGE,
	["Focus"] = FOCUS,
	["Energy"] = ENERGY,
	["Mana"] = MANA,
	
	["PvP"] = PVP,
	
	-- genders
	["Male"] = MALE,
	["Female"] = FEMALE,
	
	-- shortgenders
	["m"] = "m",
	["f"] = "f",
	
	["Rested"] = "Rested",
	
	["Leader"] = "Leader",
	["L"] = "L", -- short leader
}
if GetLocale() == "deDE" then
	-- alter L here
end

local RAID_CLASS_COLORS = RAID_CLASS_COLORS
local UNKNOWN = UNKNOWN or "Unknown"
local unpack = unpack

local new, del
do
	local list = setmetatable({}, {__mode='k'})
	function new(...)
		local t = next(list)
		if t then
			list[t] = nil
			for i = 1, select('#', ...) do
				t[i] = select(i, ...)
			end
			return t
		else
			return {...}
		end
	end
	function del(t)
		for k in pairs(t) do
			t[k] = nil
		end
		list[t] = true
		return nil
	end
end

local colors = {}
for class, data in pairs(RAID_CLASS_COLORS) do
	colors[class] = { data.r, data.g, data.b, }
end
colors.unknown = { 0.8, 0.8, 0.8 }
colors.hostile = { 226/255, 45/255, 75/255 }
colors.neutral = { 1, 1, 34/255 }
colors.friendly = { 0.2, 0.8, 0.15 }
colors.civilian = { 48/255, 113/255, 191/255 }
colors.tapped = { 0.5, 0.5, 0.5 }
colors.dead = { 0.6, 0.6, 0.6 }
colors.disconnected = { 0.7, 0.7, 0.7 }
colors.inCombat = { 1, 0, 0 }
colors.resting = { 1, 1, 0 }
colors.petHappy = { 0, 1, 0 }
colors.petNeutral = { 1, 1, 0 }
colors.petAngry = { 1, 0, 0 }
colors.rage = { 226/255, 45/255, 75/255 }
colors.energy = { 1, 220/255, 25/255 }
colors.focus = { 1, 210/255, 0 }
colors.mana = { 48/255, 113/255, 191/255 }

local function GetClassColor(unit)
	local _,class = UnitClass(unit)
	return unpack(colors[class] or colors.unknown)
end

local function GetHostilityColor(unit)
	if UnitIsPlayer(unit) or UnitPlayerControlled(unit) then
		if UnitCanAttack(unit, "player") then
			-- they can attack me
			if UnitCanAttack("player", unit) then
				-- and I can attack them
				return unpack(colors.hostile)
			else
				-- but I can't attack them
				return unpack(colors.civilian)
			end
		elseif UnitCanAttack("player", unit) then
			-- they can't attack me, but I can attack them
			return unpack(colors.neutral)
		elseif UnitIsPVP(unit) then
			-- on my team
			return unpack(colors.friendly)
		else
			-- either enemy or friend, no violence
			return unpack(colors.civilian)
		end
	elseif (UnitIsTapped(unit) and not UnitIsTappedByPlayer(unit)) or UnitIsDead(unit) then
		return unpack(colors.tapped)
	else
		local reaction = UnitReaction(unit, "player")
		if reaction then
			if reaction >= 5 then
				return unpack(colors.friendly)
			elseif reaction == 4 then
				return unpack(colors.neutral)
			else
				return unpack(colors.hostile)
			end
		else
			return unpack(colors.unknown)
		end
	end
end

-- Note, these tables use the unlocalized second returns of UnitClass and UnitRace, thus the keys are not localized.
local shortraces = {
	BloodElf = L["BE"],
	Draenei = L["Dr"],
	Dwarf = L["Dw"],
	Gnome = L["Gn"],
	Human = L["Hu"],
	NightElf = L["NE"],
	Orc = L["Or"],
	Tauren = L["Ta"],
	Troll = L["Tr"],
	Undead = L["Ud"],
}
local shortclasses = {
	WARRIOR = L["Wr"],
	PRIEST = L["Pr"],
	MAGE = L["Ma"],
	SHAMAN = L["Sh"],
	PALADIN = L["Pa"],
	WARLOCK = L["Wl"],
	DRUID = L["Dr"],
	ROGUE = L["Ro"],
	HUNTER = L["Hu"],
}

local unit -- this is set before the functions are called

local Modifiers; Modifiers = {
	-- when writing modifiers, inputs can be a number or a string. Also, any of these can be returned, as well as nil, which means a blank string.
	-- the modifier itself can be in the form of a function or a table containing a function and an event string (or a table of event strings). The events specified will cause the string to update upon that event.
	-- note: can take an argument, either nil, a number, a string.
	isequal = {function(arg, value)
		if value == arg then
			return value
		end
	end, 1},
	
	isless = {function(arg, value)
		if type(value) == type(arg) then
			if value < arg then
				return value
			end
		else
			if tostring(value) < tostring(arg) then
				return value
			end
		end
	end, 1},
	
	islessequal = {function(arg, value)
		if type(value) == type(arg) then
			if value <= arg then
				return value
			end
		else
			if tostring(value) <= tostring(arg) then
				return value
			end
		end
	end, 1},
	
	isgreater = "~islessequal",
	isgreaterequal = "~isless",
	isnotequal = "~isequal",
	
	hide = "~isequal",
	hideless = "~isless",
	hidegreater = "~isgreater",
	hidelessequal = "~islessequal",
	hidegreaterequal = "~isgreaterequal",
	hideunless = "~isnotequal",
	hidezero = "hide(0)",
	
	prepend = {function(arg, value)
		return arg .. value
	end, 1},

	append = {function(arg, value)
		return value .. arg
	end, 1},
	
	negate = function(value)
		if type(value) == "number" then
			return -value
		end
		return value
	end,
	
	percent = function(value)
		if type(value) == "number" then
			return value .. "%"
		end
		return value
	end,
	
	hidemaxhp = { function(value)
		local maxValue = UnitHealthMax(unit)
		local currValue = UnitHealth(unit)
		if maxValue == 0 or currValue == maxValue then
			return
		else
			return value
		end
	end, "UNIT_HEALTH", "UNIT_MAXHEALTH" },
	
	hidemaxmana = { function(value)
		if UnitMana(unit) == UnitManaMax(unit) then
			return
		end
		return value
	end, "UNIT_MANA", "UNIT_RAGE", "UNIT_FOCUS", "UNIT_ENERGY", "UNIT_MAXMANA", "UNIT_MAXRAGE", "UNIT_MAXFOCUS", "UNIT_MAXENERGY","UNIT_DISPLAYPOWER" },
	
	hidenomana = { function(value)
		if UnitManaMax(unit) == 0 then
			return
		end
		return value
	end, "UNIT_MANA", "UNIT_RAGE", "UNIT_FOCUS", "UNIT_ENERGY", "UNIT_MAXMANA", "UNIT_MAXRAGE", "UNIT_MAXFOCUS", "UNIT_MAXENERGY","UNIT_DISPLAYPOWER" },
	
	hidenocombat = { function(value)
		if UnitAffectingCombat(unit) then
			return value
		end
	end, "PLAYER_REGEN_ENABLED", "PLAYER_REGEN_DISABLED", "Update" },
	
	hidegone = {function(value)
		if UnitIsDeadOrGhost(unit) or not UnitIsConnected(unit) then
			return nil
		end
		return value
	end, "PLAYER_ALIVE", "PLAYER_DEAD", "PLAYER_UNGHOST" },
	
	short = function(value)
		if type(value) ~= "number" then
			if value and value:match("/") then
				local a,b = value:match("^(%d+)/(%d+)$")
				a = Modifiers.short(tonumber(a))
				b = Modifiers.short(tonumber(b))
				if a and b then
					return a.."/"..b
				end
			end
			return value
		end
		if value >= 10000000 or value <= -10000000 then
			return ("%.1fm"):format(value / 1000000)
		elseif value >= 1000000 or value <= -1000000 then
			return ("%.2fm"):format(value / 1000000)
		elseif value >= 100000 or value <= -100000 then
			return ("%.0fk"):format(value / 1000)
		elseif value >= 10000 or value <= -10000 then
			return ("%.1fk"):format(value / 1000)
		else
			return value
		end
	end,
	
	veryshort = function(value)
		if type(value) ~= "number" then
			if value and value:match("/") then
				local a,b = value:match("^(%d+)/(%d+)$")
				a = Modifiers.veryshort(tonumber(a))
				b = Modifiers.veryshort(tonumber(b))
				if a and b then
					return a.."/"..b
				end
			end
			return value
		end
		if value >= 1000000 or value <= -1000000 then
			return ("%.0fm"):format(value / 1000000)
		elseif value >= 1000 or value <= -1000 then
			return ("%.0fk"):format(value / 1000)
		else
			return value
		end
	end,
	
	upper = function(value)
		if type(value) == "string" then
			return value:upper()
		end
		return value
	end,
	
	lower = function(value)
		if type(value) == "string" then
			return value:lower()
		end
		return value
	end,
	
	bracket = function(value)
		return ("[%s]"):format(value)
	end,
	
	paren = function(value)
		return ("(%s)"):format(value)
	end,
	
	angle = function(value)
		return ("<%s>"):format(value)
	end,
	
	brace = function(value)
		return ("{%s}"):format(value)
	end,
	
	trunc = function(arg, value)
		if not arg then
			return value
		end
		return tostring(value):sub(1, arg)
	end,
	
	statuscolor = {function(value)
		local r,g,b
		if not UnitIsConnected(unit) then
			r,g,b = unpack(colors.disconnected)
		elseif UnitIsDeadOrGhost(unit) then
			r,g,b = unpack(colors.dead)
		elseif UnitAffectingCombat(unit) then
			r,g,b = unpack(colors.inCombat)
		elseif u == "player" and IsResting() then
			r,g,b = unpack(colors.resting)
		else
			if value then
				return ("|r%s"):format(value)
			else
				return "|r"
			end
		end	
		if value then
			return ("|cff%02x%02x%02x%s|r"):format(r*255, g*255, b*255, value)
		else
			return ("|cff%02x%02x%02x"):format(r*255, g*255, b*255)
		end
	end, "Update" --[[due to the combat thing]]},
	
	happycolor = { function(value)
		local x = GetPetHappiness()
		local r,g,b
		if x == 3 then
			r,g,b = unpack(colors.petHappy)
		elseif x == 2 then
			r,g,b = unpack(colors.petNeutral)
		elseif x == 1 then
			r,g,b = unpack(colors.petAngry)
		else
			return value
		end
		if value then
			return ("|cff%02x%02x%02x%s|r"):format(r*255, g*255, b*255, value)
		else
			return ("|cff%02x%02x%02x"):format(r*255, g*255, b*255)
		end
	end, "UNIT_HAPPINESS" },

	hostilecolor = { function(value)
		local r, g, b = GetHostilityColor(unit)
		if value then
			return ("|cff%02x%02x%02x%s|r"):format(r*255, g*255, b*255, value)
		else
			return ("|cff%02x%02x%02x"):format(r*255, g*255, b*255)
		end
	end, "UNIT_FACTION" },
	
	classcolor = function(value)
		local r, g, b = GetClassColor(unit)
		if value then
			return ("|cff%02x%02x%02x%s|r"):format(r*255, g*255, b*255, value)
		else
			return ("|cff%02x%02x%02x"):format(r*255, g*255, b*255)
		end
	end,
	
	hpcolor = { function(value)
		local r, g, b
		b = 0
		local perc = UnitHealth(unit) / UnitHealthMax(unit)
		if perc > 0.5 then
			g = 1
			r = 2 * (1 - perc)
		else
			g = perc * 2
			r = 1
		end
		if value then
			return ("|cff%02x%02x%02x%s|r"):format(r*255, g*255, b*255, value)
		else
			return ("|cff%02x%02x%02x"):format(r*255, g*255, b*255)
		end
	end, "UNIT_HEALTH", "UNIT_MAXHEALTH" },
	
	aggrocolor = { function(value)
		local val = UnitReaction("player", unit) or 5
	
		local info = UnitReactionColor[val]
		if value then
			return ("|cff%02x%02x%02x%s|r"):format(info.r * 255, info.g * 255, info.b * 255, value)
		else
			return ("|cff%02x%02x%02x"):format(info.r * 255, info.g * 255, info.b * 255)
		end
	end, "UNIT_FACTION" },
	
	difficultycolor = { function(value)
		if UnitCanAttack("player", unit) then
			local level = UnitLevel(unit)
			if level == 0 then
				level = 99
			end
			local info = GetDifficultyColor(level)
			if value then
				return ("|cff%02x%02x%02x%s|r"):format(info.r * 255, info.g * 255, info.b * 255, value)
			else
				return ("|cff%02x%02x%02x"):format(info.r * 255, info.g * 255, info.b * 255)
			end
		end
		return value
	end, "UNIT_LEVEL", "PLAYER_LEVEL_UP" },
	
	manacolor = { function(value)
		local powerType = UnitPowerType(unit)
		local r,g,b
		if powerType == 0 then
			r,g,b = unpack(colors.mana)
		elseif powerType == 1 then
			r,g,b = unpack(colors.rage)
		elseif powerType == 2 then
			r,g,b = unpack(colors.focus)
		elseif powerType == 3 then
			r,g,b = unpack(colors.energy)
		else
			r,g,b = unpack(colors.unknown)
		end
		if value then
			return ("|cff%02x%02x%02x%s|r"):format(r * 255, g * 255, b * 255, value)
		else
			return ("|cff%02x%02x%02x"):format(r * 255, g * 255, b * 255)
		end
	end, "UNIT_DISPLAYPOWER" },
	
	combatcolor = { function(value)
		if UnitAffectingCombat(unit) then
			local r,g,b = unpack(colors.inCombat)
			if value then
				return ("|cff%02x%02x%02x%s|r"):format(r * 255, g * 255, b * 255, value)
			else
				return ("|cff%02x%02x%02x"):format(r * 255, g * 255, b * 255)
			end
		else
			if value then
				return ("|r%s"):format(value)
			else
				return "|r"
			end
		end
	end, "PLAYER_REGEN_ENABLED", "PLAYER_REGEN_DISABLED", "Update" },
	
	color = {function(arg, value)
		if value then
			return ("|cff%s%s|r"):format(arg, value)
		else
			return ("|cff%s"):format(arg)
		end
	end, 1},
	-- colors
	white = "color(ffffff)",
	red = "color(ff0000)",
	green = "color(00ff00)",
	blue = "color(0000ff)",
	cyan = "color(00ffff)",
	fuchsia = "color(ff00ff)",
	yellow = "color(ffff00)",
	gray = "color(afafaf)",
}

local Tags = {
	-- When writing tags, `unit` is modified before the call, that contains the unit id.
	-- Also, nil (which translates to empty string), a number, or a string can be returned.
	-- the tag itself can be in the form of a function or a table containing a function and an event string (or a table of event strings). The events specified will cause the string to update upon that event.
	-- Note: can take an argument, either nil, a number, or a string.
	
	isfriend = { function()
		return UnitIsFriend("player", unit) and "true"
	end, "UNIT_FACTION"},
	
	isenemy = "~isfriend",
	
	text = {function(arg)
		return arg
	end, 1},
	
	name = { function()
		local name = UnitName(unit)
		if not name then
			return unit
		end
		return name
	end, "UNIT_NAME_UPDATE"},
	
	namehostile = "name:hostilecolor",
	
	status = {function()
		if not UnitIsConnected(unit) then
			return L["Offline"]
		elseif UnitIsDead(unit) then
			return L["Dead"]
		elseif UnitIsGhost(unit) then
			return L["Ghost"]
		elseif UnitAffectingCombat(unit) then
			return L["Combat"]
		elseif unit == "player" and IsResting() then
			return L["Resting"]
		end
	end, "Update" --[[due to the combat thing]] },
	
	statusgone = {function()
		if not UnitIsConnected(unit) then
			return L["Offline"]
		elseif UnitIsDead(unit) then
			return L["Dead"]
		elseif UnitIsGhost(unit) then
			return L["Ghost"]
		end
	end, "PLAYER_ALIVE", "PLAYER_DEAD", "PLAYER_UNGHOST" },

	curhp = { function()
		local currValue = UnitHealth(unit)
		local MobHealth3 = MobHealth3
		if MobHealth3 and not UnitIsFriend("player", unit) then
			local maxValue = UnitHealthMax(unit)
			local curr, max, MHfound = MobHealth3:GetUnitHealth(unit, currValue, maxValue)
			if MHfound then
				return curr
			end
		end
		return currValue or 0
	end, "UNIT_HEALTH", "UNIT_MAXHEALTH" },
	
	maxhp = { function()
		local currValue, maxValue = UnitHealth(unit), UnitHealthMax(unit)
		if maxValue == 0 then
			return 0
		end
		local MobHealth3 = MobHealth3
		if MobHealth3 and not UnitIsFriend("player", unit) then
			local curr, max, MHfound = MobHealth3:GetUnitHealth(unit, currValue,maxValue)
			if MHfound then
				return max
			end
		end
		return maxValue
	end, "UNIT_HEALTH", "UNIT_MAXHEALTH" },
	
	percenthp = { function()
		local hpmax = UnitHealthMax(unit)
		if hpmax == 0 then
			return 0
		end
		return ("%.0f"):format(UnitHealth(unit) / hpmax * 100)
	end, "UNIT_HEALTH", "UNIT_MAXHEALTH" },
	
	missinghp = { function()
		local currValue, maxValue = UnitHealth(unit), UnitHealthMax(unit)
		if maxValue == 0 then
			return 0
		end
		local MobHealth3 = MobHealth3
		if MobHealth3 and not UnitIsFriend("player", unit) then
			local curr, max, MHfound = MobHealth3:GetUnitHealth(unit, currValue, maxValue)
			if MHfound then
				return max - curr
			end
		end
		return maxValue - currValue
	end, "UNIT_HEALTH", "UNIT_MAXHEALTH" },
	
	fractionalhp = { function()
		local currValue, maxValue = UnitHealth(unit), UnitHealthMax(unit)
		local MobHealth3 = MobHealth3
		if maxValue > 100 then
			return currValue.."/"..maxValue
		elseif MobHealth3 and not UnitIsFriend("player", unit) then
			local curr, max, MHfound = MobHealth3:GetUnitHealth(unit, currValue, maxValue)
			if MHfound then
				return curr.."/"..max
			end
		end
		return currValue.."/"..maxValue
	end, "UNIT_HEALTH", "UNIT_MAXHEALTH" },
	
	smarthp = { function()
		local currValue, maxValue = UnitHealth(unit), UnitHealthMax(unit)
		local MobHealth3 = MobHealth3
		if maxValue > 100 then
			return currValue.."/"..maxValue
		end
		if MobHealth3 and not UnitIsFriend("player", unit) then
			local curr, max, MHfound = MobHealth3:GetUnitHealth(unit, currValue, maxValue)
			if MHfound then
				return curr.."/"..max
			end
		end
		return ("%.0f%%"):format(currValue / maxValue * 100)
	end, "UNIT_HEALTH", "UNIT_MAXHEALTH" },

	curmana = { function()
		return UnitMana(unit) or 0
	end, "UNIT_MANA", "UNIT_RAGE", "UNIT_FOCUS", "UNIT_ENERGY", "UNIT_MAXMANA", "UNIT_MAXRAGE", "UNIT_MAXFOCUS", "UNIT_MAXENERGY","UNIT_DISPLAYPOWER" },
	
	maxmana = { function()
		return UnitManaMax(unit)
	end, "UNIT_MANA", "UNIT_RAGE", "UNIT_FOCUS", "UNIT_ENERGY", "UNIT_MAXMANA", "UNIT_MAXRAGE", "UNIT_MAXFOCUS", "UNIT_MAXENERGY","UNIT_DISPLAYPOWER" },
	
	percentmana = { function()
		local mpmax = UnitManaMax(unit)
		if mpmax == 0 then
			return 0
		end
		return ("%.0f"):format(UnitMana(unit) / mpmax * 100)
	end, "UNIT_MANA", "UNIT_RAGE", "UNIT_FOCUS", "UNIT_ENERGY", "UNIT_MAXMANA", "UNIT_MAXRAGE", "UNIT_MAXFOCUS", "UNIT_MAXENERGY","UNIT_DISPLAYPOWER" },
	
	missingmana = { function()
		return UnitManaMax(unit) - UnitMana(unit)
	end, "UNIT_MANA", "UNIT_RAGE", "UNIT_FOCUS", "UNIT_ENERGY", "UNIT_MAXMANA", "UNIT_MAXRAGE", "UNIT_MAXFOCUS", "UNIT_MAXENERGY","UNIT_DISPLAYPOWER" },
	
	fractionalmana = { function()
		return UnitMana(unit).."/"..UnitManaMax(unit)
	end, "UNIT_MANA", "UNIT_RAGE", "UNIT_FOCUS", "UNIT_ENERGY", "UNIT_MAXMANA", "UNIT_MAXRAGE", "UNIT_MAXFOCUS", "UNIT_MAXENERGY","UNIT_DISPLAYPOWER" },
	
	typemana = { function()
		local p = UnitPowerType(unit)
		if p == 1 then
			return L["Rage"]
		elseif p == 2 then
			return L["Focus"]
		elseif p == 3 then
			return L["Energy"]
		else
			return L["Mana"]
		end
	end, "UNIT_DISPLAYPOWER" },
	
	level = { function()
		local x = UnitLevel(unit)
		if x > 0 then
			return x
		else
			return "??"
		end
	end, "UNIT_LEVEL" },
	
	class = function()
		return UnitClass(unit) or UNKNOWN
	end,
	
	smartclass = function()
		if UnitIsPlayer(unit) or (not UnitIsFriend("player", unit) and not UnitPlayerControlled(unit)) then
			return UnitClass(unit)
		end
	end,
	
	playerclass = function()
		if UnitIsPlayer(unit) then
			return UnitClass(unit)
		end
	end,
	
	shortclass = function()
		local _,c = UnitClass(unit)
		if c then
			return shortclasses[c]
		end
	end,
	
	shortsmartclass = function()
		if UnitIsPlayer(unit) or (not UnitIsFriend("player", unit) and not UnitPlayerControlled(unit)) then
			local _,c = UnitClass(unit)
			return shortclasses[c]
		end
	end,
	
	shortplayerclass = function()
		if UnitIsPlayer(unit) then
			local _,c = UnitClass(unit)
			return shortclasses[c]
		end
	end,
	
	creature = function()
		return UnitCreatureFamily(unit) or UnitCreatureType(unit) or UNKNOWN
	end,
	
	combos = { function()
		return GetComboPoints() or 0
	end, "PLAYER_COMBO_POINTS" },
	
	combosymbols = { function()
		return ("@"):rep(GetComboPoints())
	end, "PLAYER_COMBO_POINTS" },
	
	classification = function()
		if UnitClassification(unit) == "rare" then
			return L["Rare"]
		elseif UnitClassification(unit) == "rareelite" then
			return L["Rare-Elite"]
		elseif UnitClassification(unit) == "elite" then
			return L["Elite"]
		elseif UnitClassification(unit) == "worldboss" then
			return L["Boss"]
		end
	end,
	
	shortclassification = function()
		if UnitClassification(unit) == "rare" then
			return L["r"]
		elseif UnitClassification(unit) == "rareelite" then
			return L["r+"]
		elseif UnitClassification(unit) == "elite" then
			return L["+"]
		elseif UnitClassification(unit) == "worldboss" then
			return L["b"]
		end
	end,
	
	faction = function()
		local _, faction = UnitFactionGroup(unit)
		return faction
	end,
	
	connect = function()
		if not UnitIsConnected(unit) then
			return L["Offline"]
		end
	end,
	
	smartrace = function()
		if UnitIsPlayer(unit) then
			return UnitRace(unit) or UNKNOWN
		else
			return UnitCreatureFamily(unit) or UnitCreatureType(unit) or UNKNOWN
		end
	end,
	
	race = function()
		return UnitRace(unit)
	end,
	
	shortrace = function()
		local _,r = UnitRace(unit)
		return shortraces[r]
	end,
	
	pvp = function()
		return UnitIsPVP(unit) and L["PvP"]
	end,
	
	plus = function()
		return UnitIsPlusMob(unit) and "+"
	end,
	
	sex = function()
		local sex = UnitSex(unit)
		if sex == 0 then
			return L["Male"]
		elseif sex == 1 then
			return L["Female"]
		end
	end,
	
	shortsex = function()
		local sex = UnitSex(unit)
		if sex == 0 then
			return L["m"]
		elseif sex == 1 then
			return L["f"]
		end
	end,
	
	rested = {function()
		if GetRestState() then
			return L["Rested"]
		end
	end, "PLAYER_XP_UPDATE"},
	
	leader = function()
		if UnitIsPartyLeader(unit) then
			return L["L"]
		end
	end,
	
	leaderlong = function()
		if UnitIsPartyLeader(unit) then
			return L["Leader"]
		end
	end,
	
	happynum = { function()
		return GetPetHappiness() or 0
	end, "UNIT_HAPPINESS" },
	
	happytext = { function()
		return _G["PET_HAPPINESS" .. (GetPetHappiness() or 0)] or ""
	end, "UNIT_HAPPINESS" },
	
	happyicon = { function()
		local happy = GetPetHappiness()
		if happy == 3 then
			return ":D"
		elseif happy == 2 then
			return ":||"
		elseif happy == 1 then
			return "B("
		end
	end, "UNIT_HAPPINESS" },
	
	curxp = { function()
		if unit == "pet" then
			return (GetPetExperience())
		end
		return UnitXP(unit)
	end, "UNIT_PET_EXPERIENCE", "PLAYER_XP_UPDATE" },
	
	maxxp = { function()
		if unit == "pet" then
			local _,xp = GetPetExperience()
			return xp
		end
		return UnitXPMax(unit)
	end, "UNIT_PET_EXPERIENCE", "PLAYER_XP_UPDATE" },
	
	percentxp = { function()
		if unit == "pet" then
			local cur,max = GetPetExperience()
			if max > 0 then
				return ("%.0f"):format(cur/xp * 100)
			else
				return 0
			end
		end
		local max = UnitXPMax(unit)
		if max > 0 then
			return ("%.0f"):format(UnitXP(unit) / max * 100)
		end
		return 0
	end, "UNIT_PET_EXPERIENCE", "PLAYER_XP_UPDATE" },
	
	missingxp = { function()
		if unit == "pet" then
			local cur, max = GetPetExperience()
			return max - cur
		end
		return UnitXPMax(unit) - UnitXP(unit)
	end, "UNIT_PET_EXPERIENCE", "PLAYER_XP_UPDATE" },

	restxp = { function()
		return unit == "player" and GetXPExhaustion() or 0
	end, "PLAYER_XP_UPDATE" },
	
	percentrestxp = { function()
		local rest = unit == "player" and GetXPExhaustion() or 0
		local max = UnitXPMax(unit)
		if max > 0 then
			return ("%.0f"):format( rest / max * 100 )
		end
		return 0
	end, "PLAYER_XP_UPDATE" },
	
	tappedbyme = { function()
		if UnitIsTappedByPlayer(unit) then
			return "*"
		end
	end, "Update" },

	istapped = { function()
		if UnitIsTapped(unit) and not UnitIsTappedByPlayer(unit) then
			return L["Tapped"]
		end
	end, "Update" },
	
	fkey = function()
		local fkey
		if unit == "player" then
			fkey = 0
		else
			fkey = unit:match("^party(%d)$")
		end
		if fkey then
			return "F" .. (fkey + 1)
		end
	end,
	
	raidgroup = function()
		for i = 1, GetNumRaidMembers() do
			local name, rank, subgroup = GetRaidRosterInfo(i)
			if name == UnitName(unit) then
				return subgroup
			end
		end
	end,
	
	guild = function()
		return GetGuildInfo(unit)
	end,
	
	guildrank = function()
		local _, rank = GetGuildInfo(unit)
		return rank
	end,

	target = { function()
		local u = unit .. "target"
		if UnitExists(u) then
			return UnitName(u)
		end
	end, "UNIT_TARGET" },

	statuscolor = Modifiers.statuscolor,
	happycolor = Modifiers.happycolor,
	hostilecolor = Modifiers.hostilecolor,
	classcolor = Modifiers.classcolor,
	hpcolor = Modifiers.hpcolor,
	aggrocolor = Modifiers.aggrocolor,
	difficultycolor = Modifiers.difficultycolor,
	manacolor = Modifiers.manacolor,
	combatcolor = Modifiers.combatcolor,
	color = Modifiers.color,
	-- colors
	white = Modifiers.white,
	red = Modifiers.red,
	green = Modifiers.green,
	blue = Modifiers.blue,
	cyan = Modifiers.cyan,
	fuchsia = Modifiers.fuchsia,
	yellow = Modifiers.yellow,
	gray = Modifiers.gray,
	
	smartmissinghp = "missinghp:hidezero:negate"
}

local function toliteral(value)
	local num = tonumber(value)
	if num and not value:match("^0[^%.]") then
		return ("%s"):format(num)
	else
		return ("%q"):format(value)
	end
end

local function splitCode(code)
	local alpha, bravo, charlie = code:match("^([^:!%?]*)([:!%?])(.*)$")
	if not alpha then
		return code
	end
	return alpha, bravo, splitCode(charlie)
end

local function splitParam(code)
	local negate, alpha, bravo = code:match("^(~*)([a-zA-Z]+)%(([^%)]+)%)$")
	if alpha then
		return alpha, bravo, ((negate:len() % 2) == 1 and true or false)
	else
		negate, alpha = code:match("^(~*)(.+)$")
		if alpha then
			return alpha, nil, ((negate:len() % 2) == 1 and true or false)
		else
			return code, nil, nil
		end
	end
end

local function compileTree(code)
	if type(code) == "string" then
		return code, 0
	end
	local ends
	code[1], ends = compileTree(code[1])
	if not code[1] then
		code = del(code)
		return nil
	end
	local t = new()
	if not code[2] then
		local tag, param, negate = splitParam(code[1])
		
		local func = Tags[tag]
		if func then
			if param then
				if type(func) ~= "table" or type(func[2]) ~= "number" then
					t = del(t)
					code = del(code)
					return nil
				end
			else
				if type(func) == "table" and type(func[2]) == "number" then
					t = del(t)
					code = del(code)
					return nil
				end
			end
			t[#t+1] = "value = "
			if negate then
				t[#t+1] = "not "
			end
			t[#t+1] = ("Tags[%q]"):format(tag)
			if type(func) == "table" then
				t[#t+1] = "[1]"
			end
			t[#t+1] = "("
			if param then
				t[#t+1] = toliteral(param)
			end
			t[#t+1] = ")"
			if negate then
				t[#t+1] = " and 'true'"
			end
			t[#t+1] = ";"
		elseif tag == "" then
			t[#t+1] = "value = "
			if negate then
				t[#t+1] = "'true'"
			else
				t[#t+1] = "nil"
			end
			t[#t+1] = ";"
		else
			t = del(t)
			code = del(code)
			return nil
		end
	elseif code[2] == "?" then
		t[#t+1] = code[1]
		t[#t+1] = ("if value then ")
		local q, e = compileTree(code[3])
		if not q then
			t = del(t)
			code = del(code)
			return nil
		end
		t[#t+1] = q
		for i = 1, e do
			t[#t+1] = "end;"
		end
		if code[4] == "!" then
			t[#t+1] = ("else ")
			local q, e = compileTree(code[5])
			if not q then
				t = del(t)
				code = del(code)
				return nil
			end
			t[#t+1] = q
			for i = 1, e do
				t[#t+1] = "end;"
			end
		end
		t[#t+1] = ("end;")
	elseif code[2] == ":" then
		t[#t+1] = code[1]
		local mod, param, negate = splitParam(code[3])
		local func = Modifiers[mod] or Tags[mod]
		if func then
			if param then
				if type(func) ~= "table" or type(func[2]) ~= "number" then
					t = del(t)
					code = del(code)
					return nil
				end
			else
				if type(func) == "table" and type(func[2]) == "number" then
					t = del(t)
					code = del(code)
					return nil
				end
			end
			local isMod = func == Modifiers[mod]
			if isMod then
				t[#t+1] = "if value then "
			else
				t[#t+1] = "if not value then "
			end
			t[#t+1] = "value = "
			if negate then
				t[#t+1] = "not "
			end
			if isMod then
				t[#t+1] = "Modifiers"
			else
				t[#t+1] = "Tags"
			end
			t[#t+1] = ("[%q]"):format(mod)
			if type(func) == "table" then
				t[#t+1] = "[1]"
			end
			t[#t+1] = "("
			if param then
				t[#t+1] = toliteral(param)
			end
			if isMod then
				if param then
					t[#t+1] = ", "
				end
				t[#t+1] = "value"
			end
			t[#t+1] = ")"
			if isMod then
				if negate then
					t[#t+1] = " and value"
				end
				t[#t+1] = " end;"
			else
				if negate then
					t[#t+1] = " and 'true'"
				end
				t[#t+1] = ";"
				ends = ends + 1
			end
		else
			t = del(t)
			code = del(code)
			return nil
		end	
	else
		t = del(t)
		code = del(code)
		return nil
	end
	code = del(code)
	local s = table.concat(t)
	t = del(t)
	return s:gsub(";", "\n"), ends
end

local function select2(min, max, ...)
	if min <= max then
		return select(min, ...), select2(min+1, max, ...)
	end
end

local function parse(...)
	local n = select('#', ...)
	if n == 1 then
		return new((...))
	end
	local q_num = 0
	local q_pos
	local x_pos
	for i = 1, n do
		if select(i, ...) == "?" then
			q_num = q_num + 1
			if not q_pos then
				q_pos = i
			end
		elseif select(i, ...) == "!" then
			q_num = q_num - 1
			if q_num == 0 then
				x_pos = i
				break
			end
		end
	end
	if q_pos then
		if not x_pos then
			return new(parse(select2(1, q_pos - 1, ...)), "?", parse(select(q_pos + 1, ...)))
		else
			return new(parse(select2(1, q_pos - 1, ...)), "?", parse(select2(q_pos + 1, x_pos - 1, ...)), "!", parse(select(x_pos + 1, ...)))
		end
	end
	return new(parse(select2(1, n - 2, ...)), ":", select(n, ...))
end

local tmp = {}
local codeToFunction = setmetatable({}, {__index=function(self, code)
	local t = new()
	t[#t+1] = ([[local DogTag = AceLibrary(%q);]]):format(MAJOR_VERSION)
	t[#t+1] = [[local Tags = DogTag.__Tags;]]
	t[#t+1] = [[local Modifiers = DogTag.__Modifiers;]]
	t[#t+1] = [[local table_concat = table.concat;]]
	t[#t+1] = [[local string_gsub = string.gsub;]]
	t[#t+1] = [[local type = type;]]
	t[#t+1] = [[local tonumber = tonumber;]]
	t[#t+1] = [[local tmp = DogTag.__tmp;]]
	t[#t+1] = [[return function() ]]
	local len = code:len()
	local st = 1
	while st <= len do
		local start, fin, left, literal, inner = code:find("^(.-)(%[([^%]]+)%])", st)
		if not start then
			local rest = code:sub(st)
			if rest:len() > 0 then
				t[#t+1] = ("tmp[#tmp+1] = %q;"):format(rest)
			end
			break
		end
		st = fin+1
		if left:len() > 0 then
			t[#t+1] = ("tmp[#tmp+1] = %q;"):format(left)
		end
		
		local u = new()
		u[#u+1] = "do local value;"
		
		local q, e = compileTree(parse(splitCode(inner)))
		if not q then
			u = del(u)
			t[#t+1] = ("tmp[#tmp+1] = %q;"):format(literal)
		else
			u[#u+1] = q
			for i = 1, e do
				u[#u+1] = "end;"
			end
			u[#u+1] = [[if value then tmp[#tmp+1] = value end;]]
			u[#u+1] = [[end;]]
			local s = table.concat(u)
			u = del(u)
			t[#t+1] = s
		end
	end
	t[#t+1] = [[local s = table_concat(tmp);]]
	t[#t+1] = [[for i = 1, #tmp do tmp[i] = nil end;]]
	t[#t+1] = [[return (string_gsub(string_gsub(string_gsub(s, "^ +", ""), " +$", ""), "  +", " "));]]
	t[#t+1] = [[end;]]
	local s = table.concat(t)
	t = del(t)
	DogTag.__Tags = Tags
	DogTag.__Modifiers = Modifiers
	DogTag.__tmp = tmp
	local val = assert(loadstring(s))()
	DogTag.__Tags = nil
	DogTag.__Modifiers = nil
	DogTag.__tmp = nil
	self[code] = val
	return val
end, __mode='k'})

local codeToEventList = setmetatable({}, {__index = function(self, key)
	local t = new()
	local len = key:len()
	local st = 1
	while st <= len do
		local start, fin, left, literal, inner = key:find("^(.-)(%[([^%]]+)%])", st)
		if not start then
			break
		end
		st = fin+1
		local code = new(splitCode(inner))
		for i = 1, #code, 2 do
			local bit, param = splitParam(code[i])
			local func
			if code[i-1] ~= ":" then
				func = Tags[bit]
			else
				func = Modifiers[bit] or Tags[bit]
			end
			if type(func) == "table" then
				for j = 2, #func do
					local event = func[j]
					if j == 2 and type(event) == "number" then
						-- ignore
					elseif event then
						t[event] = true
					end
				end
			end
		end
		code = del(code)
	end
	self[key] = t
	return t
end, __mode = 'k'})

local antiAlias
do
	local function handler(literal, inner)
		local code = new(splitCode(inner))
		local real = new()
		for i = 1, #code, 2 do
			local bit, param, negate = splitParam(code[i])
			bit = bit:lower()
			local func
			if code[i-1] ~= ":" then
				func = Tags[bit]
			else
				func = Modifiers[bit] or Tags[bit]
			end
			if func then
				if i == 1 then
					real[1] = "["
				else
					real[#real+1] = code[i-1]
				end
				if negate then
					real[#real+1] = "~"
				end
				if type(func) == "string" then
					if param and func:find("[:%(%?!]") then
						if func:find("%(arg%)") then
							real[#real+1] = func:gsub("%(arg%)", "(" .. param .. ")")
						else
							code = del(code)
							real = del(real)
							return literal
						end
					else
						real[#real+1] = func
						if param then
							real[#real+1] = "("
							real[#real+1] = param
							real[#real+1] = ")"
						end
					end
				else
					real[#real+1] = bit
					if param then
						real[#real+1] = "("
						real[#real+1] = param
						real[#real+1] = ")"
					end
				end	
			else
				code = del(code)
				real = del(real)
				return literal
			end
		end
		code = del(code)
		
		real[#real+1] = ']'
		local str = table.concat(real):gsub("~~", "")
		real = del(real)
		return str
	end
	antiAlias = setmetatable({}, {__index = function(self, key)
		local value = key
		while true do
			local lastValue = value
			value = lastValue:gsub("(%[([^%]]+)%])", handler)
			if value == lastValue then
				break
			end
		end
		self[key] = value
		return value
	end, __mode= 'k'})
end

local eventData = setmetatable({}, {__index = function(self, event)
	local t = new()
	self[event] = t
	if event ~= "Update" then
		DogTag.frame:RegisterEvent(event)
	end
	return t
end})
local buckets = setmetatable({}, {__index = function(self, event)
	local t = new()
	self[event] = t
	return t
end})

local registry
local codeHash

local function scheduleFunc()
	for text in pairs(eventData.Update) do
		if text:IsShown() then
			unit = registry[text]
			text:SetText(codeToFunction[codeHash[text]]())
		end
	end
end
local function eventFunc(units, event)
	local checkUnits = next(units)
	for text in pairs(eventData[event]) do
		if text:IsShown() then
			local u = registry[text]
			if not checkUnits or units[u] then
				unit = u
				text:SetText(codeToFunction[codeHash[text]]())
			end
		end
	end
end
local NormalUnits = {player = true, pet = true, target = true, mouseover = true, focus = true}
for i = 1, 4 do
	NormalUnits['party' .. i] = true
	NormalUnits['partypet' .. i] = true
end
for i = 1, 40 do
	NormalUnits['raid' .. i] = true
end
function DogTag:RegisterFontString(text, u, code)
	self:argCheck(text, 2, "table")
	self:argCheck(u, 3, "string")
	self:argCheck(code, 4, "string")
	code = antiAlias[code]
	if codeHash[text] then
		if codeHash[text] == code and registry[text] == u then
			if text:IsShown() then
				unit = u
				text:SetText(codeToFunction[code]())
			end
			return
		end
		self:UnregisterFontString(text)
	end
	codeHash[text] = code
	registry[text] = u
	if NormalUnits[u] then
		local events = codeToEventList[code]
	
		for event in pairs(events) do
			eventData[event][text] = true
		end
	end
	registry[text] = u
	
	if text:IsShown() then
		unit = u
		text:SetText(codeToFunction[code]())
	end
end

function DogTag:UnregisterFontString(text)
	self:argCheck(text, 2, "table")
	local code = codeHash[text]
	if not code then
		return
	end
	
	local unit = registry[text]
	registry[text] = nil
	codeHash[text] = nil
	if NormalUnits[unit] then
		local events = codeToEventList[code]
	
		for event in pairs(events) do
			local t = eventData[event]
			assert(t)
			t[text] = nil
			if not next(t) then
				eventData[event] = del(t)
				if event ~= "Update" then
					self.frame:UnregisterEvent(event)
					buckets[event] = del(buckets[event])
				end
			end
		end
	end
	
	text:SetText("")
end

function DogTag:UpdateFontString(text)
	self:argCheck(text, 2, "table")
	if registry[text] and text:IsShown() then
		unit = registry[text]
		text:SetText(codeToFunction[codeHash[text]]())
	end
end

function DogTag:UpdateAllForUnit(u)
	self:argCheck(u, 2, "string")
	
	for text, un in pairs(registry) do
		if u == un and text:IsShown() then
			unit = u
			text:SetText(codeToFunction[codeHash[text]]())
		end
	end
end

function DogTag:Evaluate(u, code)
	unit = u
	return codeToFunction[antiAlias[code]]()
end

function DogTag:PARTY_MEMBERS_CHANGED()
	for text, u in pairs(registry) do
		if text:IsShown() then
			unit = u
			text:SetText(codeToFunction[codeHash[text]]())
		end
	end
end
DogTag.PARTY_LEADER_CHANGED = DogTag.PARTY_MEMBERS_CHANGED
DogTag.RAID_ROSTER_UPDATE = DogTag.PARTY_MEMBERS_CHANGED
DogTag.PARTY_MEMBER_ENABLE = DogTag.PARTY_MEMBERS_CHANGED
DogTag.PARTY_MEMBER_DISABLE = DogTag.PARTY_MEMBERS_CHANGED
DogTag.PARTY_ENTERING_WORLD = DogTag.PARTY_MEMBERS_CHANGED

function DogTag:PLAYER_TARGET_CHANGED()
	self:UpdateAllForUnit('target')
end

function DogTag:PLAYER_FOCUS_CHANGED()
	self:UpdateAllForUnit('focus')
end

function DogTag:UPDATE_MOUSEOVER_UNIT()
	self:UpdateAllForUnit('mouseover')
end

function DogTag:PLAYER_PET_CHANGED()
	self:UpdateAllForUnit('pet')
end

function DogTag:SetColorConstantTable(t)
	colors = t
	self:PARTY_MEMBERS_CHANGED()
end

local function activate(self, oldLib, oldDeactivate)
	DogTag = self
	
	self.frame = oldLib and oldLib.frame or CreateFrame("Frame", "DogTag10Frame")
	
	self.frame:UnregisterAllEvents()
	local normal = {
		PARTY_LEADER_CHANGED = true,
		PARTY_MEMBERS_CHANGED = true,
		RAID_ROSTER_UPDATE = true,
		PARTY_MEMBER_ENABLE = true,
		PARTY_MEMBER_DISABLE = true,
		PARTY_ENTERING_WORLD = true,
		PLAYER_TARGET_CHANGED = true,
	}
	self.frame:SetScript("OnEvent", function(this, event, arg1)
		local func = self[event]
		if func then
			func(self)
			return
		end
		buckets[event][arg1 or false] = true
	end)
	local timePassed = 0
	local num = 0
	local toUpdate = {}
	self.frame:SetScript("OnUpdate", function(this, elapsed)
		timePassed = timePassed + elapsed
		if timePassed > 0.05 then
			timePassed = 0
			num = (num + 1) % 3
			if num == 0 then
				if eventData.Update then
					for text in pairs(eventData.Update) do
						toUpdate[text] = true
					end
				end
			end
			for event, units in pairs(buckets) do
				if units[false] then
					for text in pairs(eventData[event]) do
						toUpdate[text] = true
					end
				else
					for text in pairs(eventData[event]) do
						if units[registry[text]] then
							toUpdate[text] = true
						end
					end
				end
			end
			for text in pairs(toUpdate) do
				toUpdate[text] = nil
				local u = registry[text]
				if u and text:IsShown() then
					unit = u
					text:SetText(codeToFunction[codeHash[text]]())
				end
			end
		end
	end)
	
	registry = {}
	self.registry = registry
	codeHash = {}
	self.codeHash = codeHash
	if oldLib then
		for text, unit in pairs(oldLib.registry) do
			self:RegisterFontString(text, unit, oldLib.codeHash[text])
		end
	end
end

AceLibrary:Register(DogTag, MAJOR_VERSION, MINOR_VERSION, activate, nil, external)
